-- TFTCtl : VMOD TFT controller
-- Anders Nilsson
-- 2021-05-09
--
-- Based on the original code by Elod Gyorgy, Digilent,
-- but now without the std_logic_unsigned and std_logic_arith
-- libraries replaced with numeric_std.
--
-- The VMOD TFT has a 480x272 resolution, 24 bits of RGB colour.
-- The TFT controller has a powerup cycle before the TFT can be used,
-- as indicated by the TFTRdy signal.
-- The TFT controller houses a simple TFT motor generating Xpixel and
-- Ypixel signals for external use, and for timing of the data
-- enable (DE_O) signal when within the visible display area.
-- The control signals and the R/G/B outputs are to be directly
-- connected to the VMOD TFT.
-- The R/G/B inputs typically connects to the data output of a
-- picture memory addressed by the Xpixel and Ypixel signals.
-- The TFT controller assumes the analog backlight option is set,
-- since the backlight enable signal is not controlled in this
-- solution.


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity TFTCtl is 
  port (
    clk : IN std_logic;
    rst	: IN std_logic;
    -- Controls signals
    DE_O : out std_logic;
    CLK_O : out std_logic;
    DISP_O : out std_logic;
    VDDEN_O : out std_logic;
    -- R/G/B output
    R_O : out std_logic_vector(7 downto 0);
    G_O : out std_logic_vector(7 downto 0);          
    B_O : out std_logic_vector(7 downto 0);
    -- R/G/B input
    R_I : in std_logic_vector(7 downto 0);
    G_I : in std_logic_vector(7 downto 0);          
    B_I : in std_logic_vector(7 downto 0);
    -- X/Y current pixel position
    Xpixel : out unsigned(9 downto 0);
    Ypixel : out unsigned(8 downto 0)
    );
	
end entity;


architecture func of TFTCtl is
  constant CLOCKFREQ : natural := 100; --MHZ
  constant TRESET : natural := 1; --ms
  constant TPOWERUP : natural := 2; --ms
  constant TLEDWARMUP : natural := 200; --ms
  constant TRESET_CYCLES : natural := natural(CLOCKFREQ*TRESET*1000);
  constant TPOWERUP_CYCLES : natural := natural(CLOCKFREQ*TPOWERUP*1000);
  constant TLEDWARMUP_CYCLES : natural := natural(CLOCKFREQ*TLEDWARMUP*1000);
	
  signal waitCnt : natural range 0 to TLEDWARMUP_CYCLES;
  signal waitCntEn : std_logic;
  
  signal clkStop : std_logic;
  signal clk10 : natural range 0 to 9;

  type state_type is (stOff, stPowerUp, stLEDWarmup, stOn); 
  signal state, nstate : state_type;
  
  signal int_R, int_G, int_B : std_logic_vector(7 downto 0);
  
  signal int_DE : std_logic;

  signal Xpix : unsigned(9 downto 0);
  signal Ypix : unsigned(8 downto 0);
  signal TFTRdy : std_logic;

begin

  int_DE <= '0' when ((Xpix >= 480) or (Ypix >= 272)) else '1';

  --Wait States
  waitCntEn <= '1' when (state = stOff or state = stPowerUp or state = stLEDWarmup) and 
						  (state = nstate) else
					'0';

  --LCD & backlight power
  VDDEN_O <= '0' when (state = stOff) else '1';
  --Display On/Off signal
  DISP_O <= '0' when (state = stOff) or (state = StPowerUp) else '1';
  --Interface signals
  DE_O <= '0' when (state = stOff) or (state = stPowerUp) else int_DE;
  R_O <= (others => '0') when (state = stOff) or (state = stPowerUp) else R_I;
  G_O <= (others => '0') when (state = stOff) or (state = stPowerUp) else G_I;
  B_O <= (others => '0') when (state = stOff) or (state = stPowerUp) else B_I;   
   --Clock stop signal
  clkStop <= '1' when (state = stOff) or (state = stPowerUp) else '0';


----------------------------------------------------------------------------------
-- Startup state machine
----------------------------------------------------------------------------------	
   SYNC_PROC: process (clk)
   begin
     if rising_edge(clk) then
       if (rst='1') then
         state <= stOff;
       else
         state <= nstate;
       end if;
     end if;
   end process;	        			


   NEXT_STATE_DECODE: process (state, waitCnt,rst)
   begin
     nstate <= state;
     case (state) is
       when stOff => --wait for signals to stabilize
         if (waitCnt = TRESET_CYCLES) then
           nstate <= stPowerUp;
         end if;
       when stPowerUp => --turn power on first
         if (waitCnt = TPOWERUP_CYCLES) then
           nstate <= stLEDWarmup;
         end if;
       when stLEDWarmup => --turn on interface signals
         if (waitCnt = TLEDWARMUP_CYCLES) then
           nstate <= stOn;
         end if;
       when stOn => --stay in stOn, until reset
         if (rst = '1') then
           nstate <= stOff;
         end if;
       when others =>
         nstate <= stOff;
     end case;      
   end process;


----------------------------------------------------------------------------------
-- Wait Counter
----------------------------------------------------------------------------------	
  process(clk)
  begin
    if Rising_Edge(clk) then
      if (rst='1') or (waitCntEn = '0') then
        waitCnt <= 0;
      else
        waitCnt <= waitCnt + 1;
      end if;
    end if;
  end process;


----------------------------------------------------------------------------------
-- 10 MHz clock
----------------------------------------------------------------------------------	
  process(clk)
  begin
    if rising_edge(clk) then
      if (rst='1') or (clkStop='1') or (clk10 = 9) then
        clk10 <= 0;
      else
        clk10 <= clk10 + 1;
      end if;
    end if;
  end process;

  process(clk)
  begin
    if rising_edge(clk) then
      if (rst='1') or (clkStop='1') or (clk10 = 0) then
        CLK_O <= '0';
      elsif (clk10 = 5) then
        CLK_O <= '1';
      end if;
    end if;
  end process;

----------------------------------------------------------------------------------
-- TFT motor
----------------------------------------------------------------------------------	

  TFTRdy <= '0' when (state = stOff) or (state = stPowerUp) else '1';

  
  process(clk)
  begin
    if rising_edge(clk) then
      if (rst='1') or (TFTRdy='0') then
        Xpix <= (others => '0');
      else
        if (clk10 = 9) then
          if Xpix = 524 then
            Xpix <= (others => '0');
          else
            Xpix <= Xpix + 1;
          end if;
        end if;
      end if;
    end if;
  end process;

  Xpixel <= Xpix;

  process(clk)
  begin
    if rising_edge(clk) then
      if (rst='1') or (TFTRdy='0') then
        Ypix <= (others => '0');
      else
        if (clk10= 9) then
          if (Xpix = 524) then
            if (Ypix = 287) then
              Ypix <= (others => '0');
            else
              Ypix <= Ypix + 1;
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;
  
  Ypixel <= Ypix;
  
end architecture;


