-- TFT_TP_demo : Demo application using TFTCtl and TPCtl
-- Anders Nilsson
-- 2021-05-09
--
-- The TFT_TP_demo reads the touch-panel A/D X/Y values, converts
-- to TFT corresponding X/Y pixels and draws a red/green cross hair
-- on the TFT display where it is touched.
-- The demo also displays A/D conversion values on the 7-segment
-- display, using switches SW1:SW0 as follows:
-- SW1:SW0 = 00 => A/D X value (12 bit)
-- SW1:SW0 = 01 => A/D Y value (12 bit)
-- SW1:SW0 = 10 => Pen detection (1 bit)
-- SW1:SW0 = 11 => A/D X&Y values (11 downto 4)&(11 downto 4)


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.all;

entity TFT_TP_demo is
  port(
    clk : in std_logic;
    rst : in std_logic;
    -- TFT display
    DE_O : out std_logic;
    CLK_O : out std_logic;
    DISP_O : out std_logic;
    VDDEN_O : out std_logic;
    R_O : out std_logic_vector(7 downto 0);
    G_O : out std_logic_vector(7 downto 0);          
    B_O : out std_logic_vector(7 downto 0);
    -- TP touchpanel
    PENIRQ_I : in std_logic;
    CS_O : out std_logic;
    DIN_O : out std_logic;
    DOUT_I : in std_logic;
--    BUSY_I : in std_logic;
    DCLK_O : out std_logic;
    -- 7-seg display
    SEG : out std_logic_vector(7 downto 0);
    AN : out std_logic_vector(3 downto 0);
    -- Switches
    SW : in std_logic_vector(1 downto 0)
    );
end;


architecture func of TFT_TP_demo is

  component TFTCtl
    port(
      clk : IN std_logic;
      rst :  IN std_logic;
      -- Control signals
      DE_O : out std_logic;
      CLK_O : out std_logic;
      DISP_O : out std_logic;
      VDDEN_O : out std_logic;
      -- R/G/B output
      R_O : out std_logic_vector(7 downto 0);
      G_O : out std_logic_vector(7 downto 0);          
      B_O : out std_logic_vector(7 downto 0);
      -- R/G/B input
      R_I : in std_logic_vector(7 downto 0);
      G_I : in std_logic_vector(7 downto 0);          
      B_I : in std_logic_vector(7 downto 0);
      -- X/Y current pixel position
      Xpixel : out unsigned(9 downto 0);
      Ypixel : out unsigned(8 downto 0)
      );
  end component;

  component TPCtl
    Port (
      clk : in STD_LOGIC;
      rst : in STD_LOGIC;
      -- A/D value outputs
      X_O : out STD_LOGIC_VECTOR(11 downto 0);
      Y_O : out STD_LOGIC_VECTOR(11 downto 0);
      Z_O : out STD_LOGIC_VECTOR(11 downto 0);
      -- Pen interrupt
      PENIRQ_I : in STD_LOGIC;
      -- Control signals
      CS_O : out  STD_LOGIC;
      DIN_O : out  STD_LOGIC;
      DOUT_I : in  STD_LOGIC;
--	BUSY_I : in  STD_LOGIC;
      DCLK_O : out  STD_LOGIC;
      -- Pen active detection
      PEN : out std_logic
      );
  end component;

  component leddriver
     port(
       clk,rst : in std_logic;
       seg : out  std_logic_vector(7 downto 0);
       an : out  std_logic_vector(3 downto 0);
       value : in  std_logic_vector(15 downto 0)
       );
   end component;


  signal Xval : std_logic_vector(11 downto 0);
  signal Yval : std_logic_vector(11 downto 0);
  signal Zval : std_logic_vector(11 downto 0);    
  signal int_value : std_logic_vector(15 downto 0);
  signal int_PEN : std_logic;

  signal Xwin : unsigned(11 downto 0);
  signal Ywin : unsigned(11 downto 0);

  signal Xpixel : unsigned(9 downto 0);
  signal Ypixel : unsigned (8 downto 0);

  signal int_R, int_G, int_B : std_logic_vector(7 downto 0);
  
begin

  Inst_TFTCtl : TFTCtl
    port map(
      clk => clk,
      rst => rst,
      DE_O => DE_O,
      CLK_O => CLK_O,
      DISP_O => DISP_O,
      VDDEN_O => VDDEN_O,
      R_O => R_O,
      G_O => G_O,
      B_O => B_O,
      R_I => int_R,
      G_I => int_G,
      B_I => int_B,
      Xpixel => Xpixel,
      Ypixel => Ypixel
      );
  

  Inst_TPCtl : TPCtl
    port map(
      clk => clk,
      rst => rst,
      X_O => Xval,
      Y_O => Yval,
      Z_O => Zval,
      PENIRQ_I => PENIRQ_I,
      CS_O => CS_O,
      DIN_O => DIN_O,
      DOUT_I => DOUT_I,
      DCLK_O => DCLK_O,
      PEN => int_PEN
      );


  Inst_leddriver : leddriver
    port map(
      clk => clk,
      rst => rst,
      seg => SEG,
      an => AN,
      value => int_value
     );


  -- Draw red/green cross hair when pen is active and within window
  int_R <= (others => '1') when (Xpixel = Xwin) and int_PEN = '1' else (others => '0');
  int_G <= (others => '1') when (Ypixel = Ywin) and int_PEN = '1' else (others => '0');
  int_B <= (others => '0');


  -- Apply window function and scale A/D X values to corresponding
  -- X pixel position on TFT.
  -- Subtract lower window boundary from A/D X value, multiply with
  -- 480 and divide with window range (xF80-x080).
  -- 480 / (xF80-x080) = 1/8.

  Xwin <= (others => '0') when Xval < x"080" else
          to_unsigned(479, Xwin'length) when Xval > x"F80" else
          unsigned(unsigned(Xval)-x"080")/8;

  -- Apply window function and scale A/D Y values to corresponding
  -- Y pixel position on TFT.
  -- Subtract lower window boundary from A/D Y value, multiply with
  -- 272 and divide with window range (xF00-x0F0).
  -- 272 / (xF00-x0F0) = 1/13 (approx).

  Ywin <= x"000" when Yval < x"0F0" else
          to_unsigned(271, Ywin'length) when Yval > x"F00" else
          unsigned(unsigned(Yval)-x"0F0")/13;

  
  int_value <= "0000" & Xval when SW = "00" else
               "0000" & Yval when SW = "01" else
--               "0000" & Zval when SW = "10" else
               "000000000000000" & int_PEN when SW = "10" else
               Xval(11 downto 4) & Yval(11 downto 4);

end architecture;
